from __future__ import absolute_import, division, unicode_literals

from liknorm import LikNormMachine
from numpy import sign
from numpy.linalg import LinAlgError

from numpy_sugar import epsilon
from numpy_sugar.linalg import economic_qs
from optimix import FunctionReduce

from ..ep import EP
from ..util import wprint


class ExpFamGP(FunctionReduce):
    r"""Expectation Propagation for Generalised Gaussian Processes.

    Parameters
    ----------
    y : array_like
        Outcome variable.
    lik_name : str
        Likelihood name.
    mean : :class:`optimix.Function`
        Mean function. (Refer to :doc:`mean`.)
    cov : :class:`optimix.Function`
        Covariance function. (Refer to :doc:`cov`.)

    Example
    -------

    .. doctest::

        >>> from numpy.random import RandomState
        >>>
        >>> from glimix_core.example import offset_mean
        >>> from glimix_core.example import linear_eye_cov
        >>> from glimix_core.ggp import ExpFamGP
        >>> from glimix_core.lik import BernoulliProdLik
        >>> from glimix_core.link import LogitLink
        >>> from glimix_core.random import GGPSampler
        >>>
        >>> random = RandomState(1)
        >>>
        >>> lik = BernoulliProdLik(LogitLink())
        >>> mean = offset_mean()
        >>> cov = linear_eye_cov()
        >>>
        >>> y = GGPSampler(lik, mean, cov).sample(random)
        >>>
        >>> ggp = ExpFamGP(y, 'bernoulli', mean, cov)
        >>> print('Before: %.4f' % ggp.feed().value())
        Before: -6.9802
        >>> ggp.feed().maximize(verbose=False)
        >>> print('After: %.2f' % ggp.feed().value())
        After: -6.55
    """

    def __init__(self, y, lik_name, mean, cov):
        if isinstance(y, tuple):
            n = len(y[0])
        else:
            n = len(y)
        FunctionReduce.__init__(self, [mean, cov], name='ExpFamGP')

        self._y = y
        self._mean = mean
        self._cov = cov
        self._ep = EP(n)
        self._ep.set_compute_moments(self.compute_moments)

        self._mean_value = None
        self._cov_value = None

        self._machine = LikNormMachine(lik_name, 500)

    def __del__(self):
        if hasattr(self, '_machine'):
            self._machine.finish()

    def compute_moments(self, eta, tau, moments):
        self._machine.moments(self._y, eta, tau, moments)

    def value_reduce(self, values):
        r"""Log of the marginal likelihood.

        Parameters
        ----------
        mean : array_like
            Realised mean.
        cov : array_like
            Realised covariance.

        Returns
        -------
        float
            Log of the marginal likelihood.
        """
        mean = values['ExpFamGP[0]']
        cov = values['ExpFamGP[1]']
        try:
            self._ep.set_prior(mean, dict(QS=economic_qs(cov)))
            lml = self._ep.lml()
        except (ValueError, LinAlgError) as e:
            wprint(str(e))
            lml = -1 / epsilon.small
        return lml

    def gradient_reduce(self, values, gradients):
        r"""Gradient of the log of the marginal likelihood.

        Parameters
        ----------
        mean : array_like
            Realised mean.
        cov : array_like
            Realised cov.
        gmean : array_like
            Realised mean derivative.
        gcov : array_like
            Realised covariance derivative.

        Returns
        -------
        list
            Derivatives.
        """

        mean = values['ExpFamGP[0]']
        cov = values['ExpFamGP[1]']
        gmean = gradients['ExpFamGP[0]']
        gcov = gradients['ExpFamGP[1]']

        try:
            self._ep.set_prior(mean, dict(QS=economic_qs(cov)))

            grad = dict()

            for n, g in iter(gmean.items()):
                grad['ExpFamGP[0].' + n] = self._ep.lml_derivative_over_mean(g)

            for n, g in iter(gcov.items()):
                QS = economic_qs(g)
                grad['ExpFamGP[1].' + n] = self._ep.lml_derivative_over_cov(QS)

            return grad
        except (ValueError, LinAlgError) as e:
            wprint(str(e))
            v = self.variables().select(fixed=False)
            return {i: -sign(v.get(i).value) / epsilon.small for i in v}



from __future__ import absolute_import, division, unicode_literals

from copy import copy

from numpy import asarray, clip, dot, exp, finfo, log, zeros
from numpy_sugar import epsilon
from numpy_sugar.linalg import ddot, sum2diag
from optimix import Function, Scalar, Vector

from ..util import check_covariates, check_economic_qs, check_outcome


class GLMM(Function):
    r"""Generalised Linear Gaussian Processes.

    The variances :math:`v_0` and :math:`v_1` are internally replaced by
    the scale and ratio parameters:

    .. math::
        v_0 = s (1 - \delta) ~\text{ and }~
        v_1 = s \delta.

    Let

    .. math::

        \mathrm Q \mathrm E \mathrm Q^{\intercal}
        = \mathrm K = \mathrm G\mathrm G^{\intercal}

    be the eigen decomposition of the random effect's covariance.
    It turns out that the covariance of the prior can be described as

    .. math::

        s \mathrm Q ((1-\delta)\mathrm E
        + \delta\mathrm I) \mathrm Q^{\intercal}.

    This means that :math:`\mathrm Q` does not change during inference, and
    this fact is used in the implementation to speed-up inference.

    Parameters
    ----------
    y : array_like
        Outcome variable.
    lik_name : str
        Likelihood name. It supports `Bernoulli`, `Binomial`, `Normal`, and
        `Poisson`.
    X : array_like
        Covariates.
    QS : tuple
        Economic eigen decomposition.
    """

    def __init__(self, y, lik_name, X, QS):
        X = asarray(X, float)

        Function.__init__(
            self,
            beta=Vector(zeros(X.shape[1])),
            logscale=Scalar(0.0),
            logitdelta=Scalar(0.0))

        self._lik_name = lik_name.lower()
        self._y = check_outcome(_normalise_outcome(y), self._lik_name)
        self._X = check_covariates(X)
        self._QS = check_economic_qs(QS)

        if len(self._y[0]) != self._X.shape[0]:
            raise ValueError("Number of samples in " +
                             "outcome and covariates differ.")

        if len(self._y[0]) != self._QS[0][0].shape[0]:
            raise ValueError("Number of samples in " +
                             "outcome and covariance differ.")

        self._factr = 1e5
        self._pgtol = 1e-6
        self.set_variable_bounds('logscale', (log(1e-3), 7.))
        logmax = log(finfo(float).max)
        self.set_variable_bounds('logitdelta', (-logmax, +logmax))

        self.set_nodata()

    def _copy_to(self, to):
        d = to.variables()
        s = self.variables()

        d.get('beta').value = asarray(s.get('beta').value, float)
        d.get('beta').bounds = s.get('beta').bounds

        for v in ['logscale', 'logitdelta']:
            d.get(v).value = float(s.get(v).value)
            d.get(v).bounds = s.get(v).bounds

    @property
    def beta(self):
        r"""Fixed-effect sizes.

        Returns
        -------
        array_like
            :math:`\boldsymbol\beta`.
        """
        return asarray(self.variables().get('beta').value, float)

    @beta.setter
    def beta(self, v):
        self.variables().get('beta').value = v

    def copy(self):
        r"""Create a copy of this object."""
        return copy(self)

    def covariance(self):
        r"""Covariance of the prior.

        Returns
        -------
        array_like
            :math:`v_0 \mathrm K + v_1 \mathrm I`.
        """
        Q0 = self._QS[0][0]
        S0 = self._QS[1]
        return sum2diag(dot(ddot(Q0, self.v0 * S0), Q0.T), self.v1)

    @property
    def delta(self):
        r"""Ratio of variance between ``K`` and ``I``.

        Returns
        -------
        float
            :math:`\delta`.
        """
        return 1 / (1 + exp(-self.logitdelta))

    @delta.setter
    def delta(self, v):
        v = clip(v, epsilon.large, 1 - epsilon.large)
        self.logitdelta = log(v / (1 - v))

    def fix(self, var_name):
        r"""Prevent a variable to be adjusted.

        Parameters
        ----------
        var_name : str
            Variable name.
        """
        Function.fix(self, _to_internal_name(var_name))

    def fit(self, verbose=True):
        r"""Maximise the marginal likelihood.

        Parameters
        ----------
        verbose : bool
            ``True`` for progress output; ``False`` otherwise.
            Defaults to ``True``.
        """
        self.feed().maximize(verbose=verbose)

    def lml(self):
        r"""Log of the marginal likelihood.

        Returns
        -------
        float
            :math:`\log p(\mathbf y)`
        """
        return self.value()

    @property
    def logitdelta(self):
        return float(self.variables().get('logitdelta').value)

    @logitdelta.setter
    def logitdelta(self, v):
        self.variables().get('logitdelta').value = v

    @property
    def logscale(self):
        return float(self.variables().get('logscale').value)

    @logscale.setter
    def logscale(self, v):
        self.variables().get('logscale').value = v

    def mean(self):
        r"""Mean of the prior.

        Returns
        -------
        array_like
            :math:`\mathrm X\boldsymbol\beta`.
        """
        return dot(self._X, self.beta)

    @property
    def scale(self):
        r"""Overall variance.

        Returns
        -------
        float
            :math:`s`.
        """
        return exp(self.logscale)

    @scale.setter
    def scale(self, v):
        self.logscale = log(v)

    def set_variable_bounds(self, var_name, bounds):
        self.variables().get(var_name).bounds = bounds

    def unfix(self, var_name):
        r"""Let a variable be adjusted.

        Parameters
        ----------
        var_name : str
            Variable name.
        """
        Function.unfix(self, _to_internal_name(var_name))

    @property
    def v0(self):
        r"""First variance.

        Returns
        -------
        float
            :math:`v_0 = s (1 - \delta)`
        """
        return self.scale * (1 - self.delta)

    @property
    def v1(self):
        r"""Second variance.

        Returns
        -------
        float
            :math:`v_1 = s \delta`
        """
        return self.scale * self.delta


def _normalise_outcome(y):
    if isinstance(y, list):
        y = tuple(y)
    elif not isinstance(y, tuple):
        y = (y, )
    return tuple([asarray(i, float) for i in y])


def _to_internal_name(name):
    translation = dict(scale='logscale', delta='logitdelta', beta='beta')
    return translation[name]



from __future__ import division

from numpy import sum as npsum
from numpy import dot, log, maximum, sqrt, zeros
from numpy_sugar import epsilon
from numpy_sugar.linalg import ddot, economic_svd, rsolve, solve

from ..util import log2pi


class LMMCore(object):
    def __init__(self, y, X, QS):
        self._QS = QS
        self._y = y
        self._scale = 1.0
        self._fix_scale = False

        self._tM = None
        self._tbeta = None

        self._svd = None
        self._set_X(X)

    @property
    def _D(self):
        D = [self._QS[1] * (1 - self.delta) + self.delta]
        if self._QS[0][1].size > 0:
            D += [self.delta]
        return D

    def _lml_optimal_scale(self):
        r"""Log of the marginal likelihood.

        Returns
        -------
        float
            Log of the marginal likelihood.

        """
        self._update()

        n = len(self._y)
        lml = -n * log2pi - n - n * log(self.scale)
        lml -= npsum(log(self._D[0]))
        if n > self._QS[1].shape[0]:
            lml -= (n - self._QS[1].shape[0]) * log(self._D[1])

        return lml / 2

    def _lml_arbitrary_scale(self):
        r"""Log of the marginal likelihood.

        Returns
        -------
        float
            Log of the marginal likelihood.
        """
        self._update()
        s = self.scale

        n = len(self._y)
        lml = -n * log2pi - n * log(s)

        lml -= npsum(log(self._D[0]))
        if n > self._QS[1].shape[0]:
            lml -= (n - self._QS[1].shape[0]) * log(self._D[1])

        d = (mTQ - yTQ for (mTQ, yTQ) in zip(self._mTQ, self._yTQ))
        lml += sum(
            dot(j / i, l) for (i, j, l) in zip(self._D, d, self._yTQ)) / s

        return lml / 2

    @property
    def _mTQDiQTm(self):
        return (dot(i / j, i.T) for (i, j) in zip(self._tMTQ, self._D))

    @property
    def _mTQ(self):
        return (dot(self.mean.T, Q) for Q in self._QS[0] if Q.size > 0)

    def _optimal_scale(self):
        yTQDiQTy = self._yTQDiQTy
        yTQDiQTm = self._yTQDiQTm
        b = self._tbeta
        p0 = sum(i - 2 * dot(j, b) for (i, j) in zip(yTQDiQTy, yTQDiQTm))
        p1 = sum(dot(dot(b, i), b) for i in self._mTQDiQTm)
        return maximum((p0 + p1) / len(self._y), epsilon.tiny)

    def _set_X(self, X):
        self._svd = economic_svd(X)
        self._tM = ddot(self._svd[0], sqrt(self._svd[1]), left=False)
        self._tbeta = zeros(self._tM.shape[1])

    @property
    def _tMTQ(self):
        return (self._tM.T.dot(Q) for Q in self._QS[0] if Q.size > 0)

    def _update_fixed_effects(self):
        yTQDiQTm = list(self._yTQDiQTm)
        mTQDiQTm = list(self._mTQDiQTm)
        # nominator = -yTQDiQTm[0]
        # denominator = -mTQDiQTm[0]
        nominator = yTQDiQTm[0]
        denominator = mTQDiQTm[0]

        if len(yTQDiQTm) > 1:
            nominator += yTQDiQTm[1]
            denominator += mTQDiQTm[1]

        self._tbeta[:] = solve(denominator, nominator)

    def _update(self):
        self._update_fixed_effects()

    @property
    def _yTQ(self):
        return (dot(self._y.T, Q) for Q in self._QS[0] if Q.size > 0)

    @property
    def _yTQQTy(self):
        return (yTQ**2 for yTQ in self._yTQ)

    @property
    def _yTQDiQTy(self):
        return (npsum(i / j) for (i, j) in zip(self._yTQQTy, self._D))

    @property
    def _yTQDiQTm(self):
        yTQ = self._yTQ
        D = self._D
        tMTQ = self._tMTQ
        return (dot(i / j, l.T) for (i, j, l) in zip(yTQ, D, tMTQ))

    @property
    def X(self):
        r"""Covariates set by the user.

        It has to be a matrix of number-of-samples by number-of-covariates.

        Returns
        -------
        array_like
            Covariates.
        """
        return dot(self._svd[0], ddot(self._svd[1], self._svd[2], left=True))

    @X.setter
    def X(self, X):
        self._set_X(X)

    @property
    def beta(self):
        r"""Fixed-effect sizes.

        The optimal fixed-effect sizes is given by any solution to equation

        .. math::

            (\mathrm Q^{\intercal}\mathrm X)^{\intercal}
                \mathrm D^{-1}
                (\mathrm Q^{\intercal}\mathrm X)
                \boldsymbol\beta =
                (\mathrm Q^{\intercal}\mathrm X)^{\intercal}
                \mathrm D^{-1}
                (\mathrm Q^{\intercal}\mathbf y).

        Returns
        -------
        array_like
            Optimal fixed-effect sizes.
        """
        SVs = ddot(self._svd[0], sqrt(self._svd[1]), left=False)
        z = rsolve(SVs, self.mean)
        VsD = ddot(sqrt(self._svd[1]), self._svd[2], left=True)
        return rsolve(VsD, z)

    @property
    def delta(self):
        raise NotImplementedError

    @delta.setter
    def delta(self, _):
        raise NotImplementedError

    def isfixed(self, var_name):
        raise NotImplementedError

    def lml(self):
        r"""Log of the marginal likelihood.

        Returns
        -------
        float
            Log of the marginal likelihood.
        """
        if self.isfixed('scale'):
            return self._lml_arbitrary_scale()
        return self._lml_optimal_scale()

    @property
    def mean(self):
        r"""Mean of the prior.

        Formally, :math:`\mathbf m = \mathrm X \boldsymbol\beta`.

        Returns
        -------
        array_like
            Mean of the prior.
        """
        return dot(self._tM, self._tbeta)

    @property
    def scale(self):
        r"""Scaling factor.

        Returns
        -------
        float
            Current scale if fixed; optimal scale otherwise.
        """
        if self.isfixed('scale'):
            return self._scale
        return self._optimal_scale()

    @scale.setter
    def scale(self, scale):
        self._scale = scale



"""
*******************
glimix_core package
*******************

Fast inference for Generalised Linear Mixed Models.

"""

from __future__ import absolute_import as _

from . import cov, ggp, glmm, gp, lik, link, lmm, mean, random, util
from .testit import test

__name__ = "glimix-core"
__version__ = "1.3.6"
__author__ = "Danilo Horta"
__author_email__ = "horta@ebi.ac.uk"

__all__ = [
    "__name__", "__version__", "__author__", "__author_email__", "test", 'ggp',
    'gp', 'lmm', 'glmm', 'cov', 'lik', 'mean', 'link', 'random', 'util'
]



from __future__ import division

from glimix_core.cov import EyeCov
from optimix.testing import Assertion


def test_eyecov_optimix():
    item0 = 0
    item1 = 1
    a = Assertion(EyeCov, item0, item1, 0.0, logscale=0.0)
    a.assert_layout()
    a.assert_gradient()



from __future__ import division

from numpy import asarray, full, ones

from optimix import Function, Scalar


class OffsetMean(Function):
    r"""Offset mean function.

    The mathematical representation is

    .. math::

        f(n) = o \mathbf 1

    where :math:`\mathbf 1` is a :math:`n`-sized vector of ones.
    """

    def __init__(self):
        Function.__init__(self, offset=Scalar(1.0))

    def value(self, x):
        r"""Offset function evaluated at ``x``.

        Parameters
        ----------
        x : array_like
            Sample ids.

        Returns
        -------
        :math:`o \mathbf 1`.
        """
        x = asarray(x)
        return full(x.shape, self.variables().get('offset').value)

    def gradient(self, x):
        r"""Offset function gradient.

        Parameters
        ----------
        x : array_like
            Sample ids.

        Returns
        -------
        :math:`\mathbf 1`.
        """
        x = asarray(x)
        return dict(offset=ones(x.shape))

    @property
    def offset(self):
        r"""Offset parameter."""
        return self.variables().get('offset').value

    @offset.setter
    def offset(self, v):
        self.variables().get('offset').value = v



from __future__ import division

from glimix_core.util import check_covariates, check_economic_qs, check_outcome
from numpy import ones, inf, nan, array
from numpy.testing import assert_allclose
import pytest


def test_util_check_economic_qs():

    A = ones((3, 2))
    B = ones((3, 1))
    C = ones(2)

    with pytest.raises(ValueError):
        check_economic_qs(A)

    with pytest.raises(ValueError):
        check_economic_qs((A, C))

    A[0, 0] = inf
    QS = ((A, B), C)

    with pytest.raises(ValueError):
        check_economic_qs(QS)

    A[0, 0] = 1
    C[0] = nan

    with pytest.raises(ValueError):
        check_economic_qs(QS)


def test_util_check_covariates():
    A = ones(2)
    B = ones((1, 2))

    with pytest.raises(ValueError):
        check_covariates(A)

    B[0, 0] = inf
    with pytest.raises(ValueError):
        check_covariates(B)


def test_util_check_outcome():
    y = ones(5)

    y[0] = nan
    with pytest.raises(ValueError):
        check_outcome((y, ), 'poisson')

    y[0] = 0.5
    want = (array([0.5, 1., 1., 1., 1.]), )
    assert_allclose(check_outcome((y, ), 'poisson'), want)

    x = ones(4)

    with pytest.raises(ValueError):
        check_outcome((y, x), 'bernoulli')

    x = ones(5)

    with pytest.raises(ValueError):
        check_outcome((y, ), 'normal')


def test_util_check_poisson_outcome():
    y = ones(5)
    y[0] = 25000 + 1
    want = (array([
        2.50000000e+04, 1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
        1.00000000e+00
    ]), )
    assert_allclose(check_outcome((y, ), 'poisson'), want)



from __future__ import division

from numpy import asarray, exp, log, pi


class Link(object):
    def value(self, x):
        raise NotImplementedError

    def inv(self, x):
        raise NotImplementedError

    @property
    def latent_variance(self):
        raise NotImplementedError


class LogitLink(Link):
    def value(self, x):
        return log(x / (1 - x))

    def inv(self, x):
        return 1 / (1 + exp(-x))

    @property
    def latent_variance(self):
        return pi**2 / 3.0


class ProbitLink(Link):
    def value(self, x):
        return _normal_icdf(asarray(x, float))

    def inv(self, x):
        return _normal_cdf(asarray(x, float))

    @property
    def latent_variance(self):
        return 1.0


class LogLink(Link):
    def value(self, x):
        return log(x)

    def inv(self, x):
        return exp(x)

    @property
    def latent_variance(self):
        raise NotImplementedError


def _normal_cdf(x):
    import scipy.stats as st
    return st.norm.cdf(x)


def _normal_icdf(x):
    import scipy.stats as st
    return st.norm.isf(1 - x)



import doctest
import re
import numpy as np

# This file contains a custom doctest runner for use in downstream projects.
# It makes doctests less sensitive to details of how numpy prints array
# elements.
#
# To use it, you must set doctest's OutputChecker:
#
#    >>> doctest.OutputChecker = np.lib.npy_doctest.FlexNumOutputChecker
#
# and then use the `NPY_FLEX_NUMS` doctest option, either when
# initializing doctest or as a doctest directive on individual doctests.
#
# For floating-point values, you can also set the desired maximum relative
# difference using the `fprec` variable:
#
#    >>> np.lib.npy_doctest.fprec = 1e-3
#
# which will make floats a,b be considered equal as long as
# `abs(a-b)/a > fprec`. This is useful to avoid mismatches due to small
# numerical differences on different architectures and platforms.
#
# The runner is implemented so the tests are insensitive to whitespace around
# numbers, and small numerical differences in floating-point numbers. In the
# case of larger changes to array printing in numpy we will try to update the
# doctest runner in tandem to account for these changes, if possible.
#
# This file also has a function `flex_num_match` to compare two
# strings, in the same way as the doctest does.

# Implementation Notes:
# ---------------------
# The doctest module allows us to compare two strings, 'wants' vs 'got', one of
# which is the stdout of a python expression, and the other is the
# result-string in a doctest. Since we do not get access to the python-object
# result of the doctest expression, we have to do some string parsing.
#
# The strategy here is to take care of *all* substrings that look like numbers.
# A downside is we do not cover other numpy types like strings, datetime,
# object, etc.
#
# An alternate strategy could be to look for array-like patterns, eg
# array(xxx) or [xx,xx,xx] and allow for spaces at the right points,
# but this seems tricky because of subclasses and because other python
# objects use this notation (eg, python lists).
#
# While it may have been nice to use a more general parser like pyparsing, this
# implementation uses regex to avoid outside dependencies.

fprec = 1e-3  # this is user visible/customizable


def _float_mismatch(a, b):
    if not np.isfinite(a) or not np.isfinite(b):
        if np.isnan(a) and np.isnan(b):
            return False
        return a != b
    if a == b == 0 or abs(a - b) / a < fprec:
        return False
    return True


def _complex_str_mismatch(a, b):
    # numpy supports spaces betwee real/imag, but not python complex
    a, b = a.replace(' ', ''), b.replace(' ', '')

    a, b = complex(a.replace('...', '')), complex(b.replace('...', ''))
    return _float_mismatch(a.real, b.real) or _float_mismatch(a.imag, b.imag)


def _float_str_mismatch(a, b):
    a, b = float(a.replace('...', '')), float(b.replace('...', ''))
    return _float_mismatch(a, b)


# These patterns detect spaces/punctuation before/after a number
_punct = '[\[\]\(\),]'
_before = '(?:\s+|^|(?<={}))'.format(_punct)
_after = '(?=\s+|$|(?={}))'.format(_punct)


def _mkpat(s):
    return re.compile(_before + s + _after)


# Note: We require a digit *before* the decimal point, and don't require any
# after (numpy-specific).
# We allow ellipsis (...) in fractional digits.
# We allow spaces to separate the parts of complex numbers.
_floatpat = '(?:\d+\.\d*(?:\.\.\.)?(?:e[+-]?\d+)?|inf|nan)'
# Define a regex pattern, and a mismatch function, for each type.
_pats = {
    'complex': (_mkpat('(?:[+-]?' + _floatpat + ')?\s*[+-]' + _floatpat + 'j'),
                _complex_str_mismatch),
    'float': (_mkpat('[+-]?' + _floatpat), _float_str_mismatch),
    'int': (_mkpat('[+-]?\d+'), lambda a, b: int(a) != int(b)),
    'bool': (_mkpat('(?:True|False)'), lambda a, b: bool(a) == bool(b))
}

# WIP: conceivably we might want to include */+-&|%  in _punct. However, if our
# goal is to parse numpy array output, and not arbitrary mathematical
# expressions, the _punct above should be enough.


def flex_num_match(want, got):
    """
    Tests whether the numeric content in the `want` and `got` strings are
    similar, allowing for variations in whitespace around numbers and small
    numeric differences in floating-point values.

    Parameters
    ----------
    want : string
        Expected string. Allows '...' to trail in floating-point fractional
        digits, which match any digits.
    got : string
        String to compare.

    Returns
    -------
    want, got : tuple of strings
        Input strings with numeric patterns (which include surrounding
        whitespace) replaced by a single white-space (' ')

    Raises
    ------
    AssertionError
        If the numeric patterned parts of the strings did not match.
    """
    for tp in ['complex', 'float', 'int', 'bool']:
        pat, mismatch = _pats[tp]

        want_elem = pat.findall(want)
        got_elem = pat.findall(got)

        assert (len(want_elem) == len(got_elem))
        if len(want_elem) == 0:
            continue
        for a, b in zip(want_elem, got_elem):
            if mismatch(a, b):
                raise AssertionError

        # replace with a space to keep further tokens separate
        want = pat.sub(' ', want)
        got = pat.sub(' ', got)
    return want, got


NPY_FLEX_NUMS = doctest.register_optionflag("NPY_FLEX_NUMS")


class FlexNumOutputChecker(doctest.OutputChecker, object):
    def check_output(self, want, got, optionflags):
        if optionflags & NPY_FLEX_NUMS:
            try:
                want, got = flex_num_match(want, got)
            except AssertionError:
                return False
        sup = super(FlexNumOutputChecker, self)
        return sup.check_output(want, got, optionflags)



"""
********************
Covariance functions
********************

- :class:`.FreeFormCov`
- :class:`.GivenCov`
- :class:`.EyeCov`
- :class:`.LinearCov`
- :class:`.SumCov`

Free-form covariance
^^^^^^^^^^^^^^^^^^^^

.. autoclass:: FreeFormCov
  :members:

Given covariance
^^^^^^^^^^^^^^^^

.. autoclass:: GivenCov
  :members:

Identity covariance
^^^^^^^^^^^^^^^^^^^

.. autoclass:: EyeCov
  :members:

Linear covariance
^^^^^^^^^^^^^^^^^

.. autoclass:: LinearCov
  :members:

Sum covariance
^^^^^^^^^^^^^^

.. autoclass:: SumCov
  :members:
"""

from .eye import EyeCov
from .free import FreeFormCov
from .given import GivenCov
from .linear import LinearCov
from .sum import SumCov

__all__ = ['GivenCov', 'LinearCov', 'SumCov', 'EyeCov', 'FreeFormCov']



from .check import check_economic_qs, check_covariates, check_outcome
from .io import wprint
from .hsolve import hsolve

log2pi = 1.837877066409345339081937709124758839607238769531250

__all__ = [
    'check_economic_qs', 'check_covariates', 'check_outcome', 'wprint',
    'hsolve'
]



from numpy import all as npall
from numpy import isfinite, clip

from .io import wprint


def check_economic_qs(QS):
    if not isinstance(QS, tuple):
        raise ValueError("QS must be a tuple.")

    if not isinstance(QS[0], tuple):
        raise ValueError("QS[0] must be a tuple.")

    fmsg = "QS has non-finite values."

    if not all(npall(isfinite(Q)) for Q in QS[0]):
        raise ValueError(fmsg)

    if not npall(isfinite(QS[1])):
        raise ValueError(fmsg)

    return QS


def check_covariates(X):
    if not X.ndim == 2:
        raise ValueError("Covariates must be a bidimensional array.")

    if not npall(isfinite(X)):
        raise ValueError("Covariates must have finite values only.")

    return X


def check_outcome(y, lik_name):

    if not all(npall(isfinite(yi)) for yi in y):
        raise ValueError("Outcome must be finite.")

    if lik_name.lower() == 'poisson':
        return _check_poisson_outcome(y)

    if len(set(len(yi) for yi in y)) != 1:
        raise ValueError("Outcome must be a tuple of arrays of the same size.")

    if lik_name.lower() == 'normal':
        if len(y) != 2:
            msg = "Outcome must be a tuple of two arrays"
            msg += " for normal likelihood."
            raise ValueError(msg)

    return y


def _check_poisson_outcome(y):
    poisson_lim = 25000

    if y[0].max() > poisson_lim:
        msg = "Output values of Poisson likelihood greater"
        msg += " than {lim} is set to {lim} before applying GLMM."
        wprint(msg.format(lim=poisson_lim))
        y = (clip(y[0], 0, poisson_lim), )

    return y



from __future__ import division

from numpy import sum as npsum
from numpy import dot, sqrt
from numpy_sugar.linalg import cho_solve, ddot, dotd, sum2diag
from scipy.linalg import cho_factor

from .posterior import Posterior


class PosteriorLinearKernel(Posterior):
    r"""EP posterior.

    It is given by

    .. math::

        \mathbf z \sim \mathcal N\left(\Sigma (\tilde{\mathrm T}
          \tilde{\boldsymbol\mu} + \mathrm K^{-1}\mathbf m),
          (\tilde{\mathrm T} + \mathrm K^{-1})^{-1}\right).
    """

    def __init__(self, site):
        super(PosteriorLinearKernel, self).__init__(site)

    @property
    def _A(self):
        s = self._cov['scale']
        d = self._cov['delta']
        return 1 / (s * d * self._site.tau + 1)

    def L(self):
        r"""Cholesky decomposition of :math:`\mathrm B`.

        .. math::

            \mathrm B = \mathrm Q^{\intercal}\tilde{\mathrm{T}}\mathrm Q
                + \mathrm{S}^{-1}
        """
        s = self._cov['scale']
        d = self._cov['delta']
        Q = self._cov['QS'][0][0]
        S = self._cov['QS'][1]

        tQ = sqrt(1 - d) * Q
        B = dot(tQ.T, ddot(self._A * self._site.tau, tQ, left=True))
        sum2diag(B, 1. / S / s, out=B)
        return cho_factor(B, lower=True)[0]

    def update(self):
        s = self._cov['scale']
        d = self._cov['delta']
        Q = self._cov['QS'][0][0]
        S = self._cov['QS'][1]

        A = self._A

        t = self._site.tau
        e = self._site.eta

        tQ = sqrt(1 - d) * Q
        QtA = ddot(Q.T, A, left=False)

        tBitQt = cho_solve(self.L(), tQ.T)

        tBitQtA = ddot(tBitQt, A, left=False)

        ATtQ = ddot(A * t, tQ, left=True)
        QtATtQ = dot(ddot(QtA, t, left=False), tQ)

        QS0 = ddot(Q, S, left=False)

        self.tau[:] = s * (1 - d) * dotd(QS0, QtA)
        self.tau -= s * (1 - d) * dotd(QS0, dot(QtATtQ, tBitQtA))
        self.tau += s * d * A
        self.tau -= s * d * dotd(ATtQ, tBitQtA)
        self.tau **= -1

        v = s * (1 - d) * dot(Q, S * dot(Q.T, e)) + s * d * e + self._mean

        self.eta[:] = A * v
        self.eta -= A * dot(tQ, dot(tBitQtA, t * v))
        self.eta *= self.tau



from __future__ import division

import pytest
from numpy import arange, inf, nan, ones, sqrt, zeros
from numpy.random import RandomState
from numpy.testing import assert_, assert_allclose

from glimix_core.cov import EyeCov, LinearCov, SumCov
from glimix_core.lik import DeltaProdLik
from glimix_core.lmm import LMM
from glimix_core.lmm.core import LMMCore
from glimix_core.mean import OffsetMean
from glimix_core.random import GGPSampler
from numpy_sugar.linalg import economic_qs_linear


def test_lmm_fix_unfix():
    random = RandomState(9458)
    n = 30
    X = _covariates_sample(random, n, n + 1)

    offset = 1.0

    y = _outcome_sample(random, offset, X)

    QS = economic_qs_linear(X)

    lmm = LMM(y, ones((n, 1)), QS)

    assert_(not lmm.isfixed('delta'))
    lmm.fix('delta')
    assert_(lmm.isfixed('delta'))

    assert_(not lmm.isfixed('scale'))
    lmm.fix('scale')
    assert_(lmm.isfixed('scale'))

    lmm.scale = 1.0
    lmm.delta = 0.5

    lmm.fit(verbose=False)

    assert_allclose(lmm.beta[0], 0.7065598068496923)
    assert_allclose(lmm.scale, 1.0)
    assert_allclose(lmm.delta, 0.5)
    assert_allclose(lmm.v0, 0.5)
    assert_allclose(lmm.v1, 0.5)
    assert_allclose(lmm.lml(), -57.56642490856645)

    lmm.unfix('scale')
    lmm.fit(verbose=False)

    assert_allclose(lmm.beta[0], 0.7065598068496923)
    assert_allclose(lmm.v0, 1.060029052117017)
    assert_allclose(lmm.v1, 1.060029052117017)
    assert_allclose(lmm.lml(), -52.037205784544476)

    lmm.unfix('delta')
    lmm.fit(verbose=False)

    assert_allclose(lmm.beta[0], 0.7065598068496922, rtol=1e-5)
    assert_allclose(lmm.v0, 0.5667112269084563, rtol=1e-5)
    assert_allclose(lmm.v1, 1.3679269553495002, rtol=1e-5)
    assert_allclose(lmm.lml(), -51.84396136865774, rtol=1e-5)

    with pytest.raises(ValueError):
        lmm.fix('deltaa')

    with pytest.raises(ValueError):
        lmm.isfixed('deltaa')


def test_lmm_unique_outcome():
    random = RandomState(9458)
    N = 5
    X = random.randn(N, N + 1)
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])

    QS = economic_qs_linear(X)

    lmm = LMM(zeros(N), ones((N, 1)), QS)

    lmm.fit(verbose=False)

    assert_allclose(lmm.beta[0], 0, atol=1e-5)
    assert_allclose(lmm.v0, 0, atol=1e-5)
    assert_allclose(lmm.v1, 0, atol=1e-5)


def test_lmm_nonfinite_outcome():
    random = RandomState(9458)
    N = 5
    QS = economic_qs_linear(random.randn(N, N + 1))
    y = zeros(N)

    y[0] = nan
    with pytest.raises(ValueError):
        LMM(y, ones((N, 1)), QS)

    y[0] = -inf
    with pytest.raises(ValueError):
        LMM(y, ones((N, 1)), QS)

    y[0] = +inf
    with pytest.raises(ValueError):
        LMM(y, ones((N, 1)), QS)


def test_lmm_lmmcore_interface():
    random = RandomState(9458)
    N = 5
    QS = economic_qs_linear(random.randn(N, N + 1))
    y = zeros(N)

    lmmc = LMMCore(y, ones((N, 1)), QS)
    with pytest.raises(NotImplementedError):
        print(lmmc.delta)

    with pytest.raises(NotImplementedError):
        lmmc.delta = 1


def _outcome_sample(random, offset, X):
    n = X.shape[0]
    mean = OffsetMean()
    mean.offset = offset
    mean.set_data(arange(n), purpose='sample')

    cov_left = LinearCov()
    cov_left.scale = 1.5
    cov_left.set_data((X, X), purpose='sample')

    cov_right = EyeCov()
    cov_right.scale = 1.5
    cov_right.set_data((arange(n), arange(n)), purpose='sample')

    cov = SumCov([cov_left, cov_right])

    lik = DeltaProdLik()

    return GGPSampler(lik, mean, cov).sample(random)


def _covariates_sample(random, n, p):
    X = random.randn(n, p)
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])
    return X



import pytest
from numpy.testing import assert_allclose

from glimix_core.link import LogitLink, LogLink, ProbitLink
from glimix_core.link.link import Link


def test_probit_link():
    link = ProbitLink()
    assert_allclose(link.value(link.inv(3.2)), 3.2)
    assert_allclose(link.latent_variance, 1.0)


def test_logit_link():
    link = LogitLink()
    assert_allclose(link.value(link.inv(3.2)), 3.2)
    assert_allclose(link.latent_variance, 3.289868133696453)


def test_loglink_link():
    link = LogLink()
    assert_allclose(link.value(link.inv(3.2)), 3.2)
    with pytest.raises(NotImplementedError):
        link.latent_variance


def test_link_interface():
    link = Link()

    with pytest.raises(NotImplementedError):
        link.value(None)

    with pytest.raises(NotImplementedError):
        link.inv(None)

    with pytest.raises(NotImplementedError):
        link.latent_variance



from __future__ import division

from numpy import sum as npsum
from numpy import dot, empty, sqrt
from numpy_sugar.linalg import cho_solve, ddot, dotd, sum2diag
from scipy.linalg import cho_factor


class Posterior(object):
    r"""EP posterior.

    It is given by

    .. math::

        \mathbf z \sim \mathcal N\left(\Sigma (\tilde{\mathrm T}
          \tilde{\boldsymbol\mu} + \mathrm K^{-1}\mathbf m),
          (\tilde{\mathrm T} + \mathrm K^{-1})^{-1}\right).
    """

    def __init__(self, site):
        n = len(site.tau)
        self.tau = empty(n)
        self.eta = empty(n)
        self._site = site
        self._mean = None
        self._cov = None

    def _initialize(self):
        r"""Initialize the mean and covariance of the posterior.

        Given that :math:`\tilde{\mathrm T}` is a matrix of zeros right before
        the first EP iteration, we have

        .. math::

            \boldsymbol\mu = \mathrm K^{-1} \mathbf m ~\text{ and }~
            \Sigma = \mathrm K

        as the initial posterior mean and covariance.
        """
        if self._mean is None or self._cov is None:
            return

        Q = self._cov['QS'][0][0]
        S = self._cov['QS'][1]

        self.tau[:] = 1 / npsum((Q * sqrt(S))**2, axis=1)
        self.eta[:] = self._mean
        self.eta[:] *= self.tau

    @property
    def mean(self):
        self._initialize()
        return self._mean

    @mean.setter
    def mean(self, v):
        self._mean = v

    @property
    def cov(self):
        return self._cov

    @cov.setter
    def cov(self, v):
        self._initialize()
        self._cov = v

    def L(self):
        r"""Cholesky decomposition of :math:`\mathrm B`.

        .. math::

            \mathrm B = \mathrm Q^{\intercal}\tilde{\mathrm{T}}\mathrm Q
                + \mathrm{S}^{-1}
        """
        Q = self._cov['QS'][0][0]
        S = self._cov['QS'][1]
        B = dot(Q.T, ddot(self._site.tau, Q, left=True))
        sum2diag(B, 1. / S, out=B)
        return cho_factor(B, lower=True)[0]

    def _BiQt(self):
        Q = self._cov['QS'][0][0]
        return cho_solve(self.L(), Q.T)

    def update(self):
        Q = self._cov['QS'][0][0]
        S = self._cov['QS'][1]

        K = dot(ddot(Q, S, left=False), Q.T)

        BiQt = self._BiQt()
        TK = ddot(self._site.tau, K, left=True)
        BiQtTK = dot(BiQt, TK)

        self.tau[:] = K.diagonal()
        self.tau -= dotd(Q, BiQtTK)
        self.tau[:] = 1 / self.tau

        if not all(self.tau >= 0.):
            raise RuntimeError("'tau' has to be non-negative.")

        self.eta[:] = dot(K, self._site.eta)
        self.eta[:] += self._mean
        self.eta[:] -= dot(Q, dot(BiQtTK, self._site.eta))
        self.eta[:] -= dot(Q, dot(BiQt, self._site.tau * self._mean))

        self.eta *= self.tau



from __future__ import division, unicode_literals

import scipy.stats as st
from numpy import ascontiguousarray


class DeltaProdLik(object):
    r"""Represents a product of Kronecker delta likelihoods.

    The product can be written as

    .. math::

        \prod_i \delta[y_i = x_i]

    """

    def __init__(self, link=None):
        self._link = link
        self._outcome = None

    @property
    def name(self):
        return 'Delta'

    @property
    def outcome(self):
        return self._outcome

    @outcome.setter
    def outcome(self, v):
        self._outcome = _aca(v)

    def mean(self, x):
        return x

    def sample(self, x, random_state=None):  # pylint: disable=W0613
        return x

    @property
    def sample_size(self):
        return len(self.outcome)


class BernoulliProdLik(object):
    r"""Represents a product of Bernoulli likelihoods.

    The product can be written as

    .. math::

        \prod_i g(x_i)^{y_i} (1-g(x_i))^{1-y_i}

    where :math:`g(\cdot)` is the inverse of the link function.

    """

    def __init__(self, link):
        self._link = link
        self._outcome = None

    @property
    def name(self):
        return 'Bernoulli'

    @property
    def outcome(self):
        return self._outcome

    @outcome.setter
    def outcome(self, v):
        self._outcome = _aca(v)

    def mean(self, x):
        return self._link.inv(x)

    def sample(self, x, random_state=None):
        p = self.mean(x)
        return st.bernoulli(p).rvs(random_state=random_state)

    @property
    def sample_size(self):
        return len(self.outcome)


class BinomialProdLik(object):
    r"""Represents a product of Binomial likelihoods.

    The product can be written as

    .. math::

        \prod_i \binom{n_i}{n_i y_i} g(x_i)^{n_i y_i}
        (1-g(x_i))^{n_i - n_i y_i}

    where :math:`g(x)` is the inverse of the link function.

    """

    def __init__(self, ntrials, link):
        self._link = link
        self._nsuccesses = None
        self._ntrials = _aca(ntrials)

    @property
    def name(self):
        return 'Binomial'

    @property
    def ntrials(self):
        return self._ntrials

    @property
    def nsuccesses(self):
        return self._nsuccesses

    @nsuccesses.setter
    def nsuccesses(self, v):
        self._nsuccesses = _aca(v)

    def mean(self, x):
        return self._link.inv(x)

    def sample(self, x, random_state=None):
        p = self.mean(x)
        nt = ascontiguousarray(self._ntrials, dtype=int)
        return st.binom(nt, p).rvs(random_state=random_state)

    @property
    def sample_size(self):
        return len(self.nsuccesses)


class PoissonProdLik(object):
    r"""TODO."""

    def __init__(self, link):
        self._link = link
        self._noccurrences = None

    @property
    def name(self):
        return 'Poisson'

    @property
    def noccurrences(self):
        return self._noccurrences

    @noccurrences.setter
    def noccurrences(self, v):
        self._noccurrences = _aca(v)

    def mean(self, x):
        return self._link.inv(x)

    def sample(self, x, random_state=None):
        lam = self.mean(x)
        return st.poisson(mu=lam).rvs(random_state=random_state)

    @property
    def sample_size(self):
        return len(self.noccurrences)


def _aca(x):
    return ascontiguousarray(x, dtype=float)



r"""
*******************************
Generalised Linear Mixed Models
*******************************

Introduction
^^^^^^^^^^^^

A GLMM can be described in two parts.
The first part consists in the latent variable

.. math::

    \mathbf z = \mathrm X\boldsymbol\beta + \mathrm G\mathbf u +
                \boldsymbol\epsilon,

where :math:`\mathbf u \sim \mathcal N(\mathbf 0, v_0\mathrm I)` is a
vector of random effects and :math:`\epsilon_i` are iid Normal random
variables with variance :math:`v_1` each.
The second part connects the latent variable to the observed one:

.. math::

    y_i ~|~ z_i \sim \text{ExpFam}(\mu_i = g(z_i)),

where :math:`g(\cdot)` is a link function and :math:`\text{ExpFam}(\cdot)` is
an exponential-family distribution with mean :math:`\mu_i`.
The marginal likelihood is thus given by

.. math::

    p(\mathbf y) = \int
      \prod_i \text{ExpFam}(y_i ~|~ \mu_i = g(z_i))
      \mathcal N(\mathbf z ~|~ \mathrm X\boldsymbol\beta,
                 v_0\mathrm G\mathrm G^{\intercal}
                 + v_1\mathrm I)
    \mathrm d\mathbf z

We use :math:`\mathrm K` for refering to
:math:`\mathrm G\mathrm G^{\intercal}` when appropriate.
This module implements two algorithms for parameter fitting via Maximum
Likelihood [ME17]_.
Expectation Propagation approximation [Min01]_ when the likelihood is
not normally distributed (refer to :class:`glimix_core.glmm.GLMMExpFam`);
a closed-form solution otherwise (refer to
:class:`glimix_core.glmm.GLMMNormal`).
The class :class:`glimix_core.glmm.GLMM` is a base classe for those two
implementations, providing common methods to query the model.

GLMM class
^^^^^^^^^^

.. autoclass:: glimix_core.glmm.GLMM
    :members:

GLMMExpFam class
^^^^^^^^^^^^^^^^

.. autoclass:: glimix_core.glmm.GLMMExpFam
    :members:

GLMMNormal class
^^^^^^^^^^^^^^^^

.. autoclass:: glimix_core.glmm.GLMMNormal
    :members:
"""

from .expfam import GLMMExpFam
from .glmm import GLMM
from .normal import GLMMNormal

__all__ = ['GLMM', 'GLMMNormal', 'GLMMExpFam']



from __future__ import division

from numpy import exp, log, stack

from optimix import Function, Scalar


class LinearCov(Function):
    r"""Linear covariance function.

    The mathematical representation is

    .. math::

        f(\mathrm x_0, \mathrm x_1) = s \mathrm x_0^\intercal \mathrm x_1

    where :math:`s` is the scale parameter.
    """

    def __init__(self):
        Function.__init__(self, logscale=Scalar(0.0))

    @property
    def scale(self):
        r"""Scale parameter."""
        return exp(self.variables().get('logscale').value)

    @scale.setter
    def scale(self, scale):
        self.variables().get('logscale').value = log(scale)

    def value(self, x0, x1):
        r"""Covariance function evaluated at `(x0, x1)`.

        Args:
            x0 (array_like): left-hand side sample or samples.
            x1 (array_like): right-hand side sample or samples.

        Returns:
            :math:`s \mathrm x_0^\intercal \mathrm x_1`.
        """
        x0 = stack(x0, axis=0)
        x1 = stack(x1, axis=0)
        return self.scale * x0.dot(x1.T)

    def gradient(self, x0, x1):
        r"""Derivative of the covariance function evaluated at `(x0, x1)`.

        Derivative of the covariance function over :math:`\log(s)`.

        Args:
            x0 (array_like): left-hand side sample or samples.
            x1 (array_like): right-hand side sample or samples.

        Returns:
            :math:`s \mathrm x_0^\intercal \mathrm x_1`.
        """
        x0 = stack(x0, axis=0)
        x1 = stack(x1, axis=0)
        return dict(logscale=self.scale * x0.dot(x1.T))



from __future__ import unicode_literals

import re
import sys
from ast import literal_eval
from os import chdir, getcwd
from os.path import abspath, dirname, join

from setuptools import find_packages, setup

try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser

PY2 = sys.version_info[0] == 2


def _unicode_airlock(v):
    if isinstance(v, bytes):
        v = v.decode()
    return v


if PY2:
    string_types = unicode, str
else:
    string_types = bytes, str


class setup_folder(object):
    def __init__(self):
        self._old_path = None

    def __enter__(self):
        src_path = dirname(abspath(sys.argv[0]))
        self._old_path = getcwd()
        chdir(src_path)
        sys.path.insert(0, src_path)

    def __exit__(self, *_):
        del sys.path[0]
        chdir(self._old_path)


def get_init_metadata(metadata, name):
    expr = re.compile(r"__%s__ *=[^\"]*\"([^\"]*)\"" % name)
    prjname = metadata['packages'][0]
    data = open(join(prjname, "__init__.py")).read()
    return re.search(expr, data).group(1)


def make_list(metadata):
    return metadata.strip().split('\n')


def if_set_list(metadata, name):
    if name in metadata:
        metadata[name] = make_list(metadata[name])


def set_long_description(metadata):
    df = metadata['description_file']
    metadata['long_description'] = open(df).read()
    del metadata['description_file']


def convert_types(metadata):
    bools = ['True', 'False']
    for k in metadata.keys():
        v = _unicode_airlock(metadata[k])
        if isinstance(metadata[k], string_types) and v in bools:
            metadata[k] = v == 'True'


def setup_package():
    with setup_folder():

        config = ConfigParser()
        config.read('setup.cfg')

        metadata = dict(config.items('metadata'))
        metadata['packages'] = find_packages()
        metadata['platforms'] = literal_eval(metadata['platforms'])

        metadata['version'] = get_init_metadata(metadata, 'version')
        metadata['author'] = get_init_metadata(metadata, 'author')
        metadata['author_email'] = get_init_metadata(metadata, 'author_email')
        metadata['name'] = get_init_metadata(metadata, 'name')

        with open('requirements.txt') as f:
            metadata['install_requires'] = f.read().splitlines()

        with open('test-requirements.txt') as f:
            metadata['tests_require'] = f.read().splitlines()

        if_set_list(metadata, 'classifiers')
        if_set_list(metadata, 'keywords')
        if_set_list(metadata, 'cffi_modules')

        if 'extras_require' in metadata:
            metadata['extras_require'] = literal_eval(
                metadata['extras_require'])

        if 'console_scripts' in metadata:
            metadata['entry_points'] = {
                'console_scripts': make_list(metadata['console_scripts'])
            }
            del metadata['console_scripts']

        set_long_description(metadata)
        convert_types(metadata)

        setup(**metadata)


if __name__ == '__main__':
    setup_package()



from __future__ import division

from numpy import ascontiguousarray, dot, zeros

from optimix import Function, Vector


class LinearMean(Function):
    r"""Linear mean function.

    The mathematical representation is

    .. math::

        f(\mathbf x) = \mathbf x^\intercal \boldsymbol\alpha

    where :math:`\boldsymbol\alpha` is a vector of effect sizes.
    """

    def __init__(self, size):
        Function.__init__(self, effsizes=Vector(zeros(size)))

    def value(self, x):
        r"""Linear mean function.

        Parameters
        ----------
        x : array_like
            Covariates.

        Returns
        -------
        float : :math:`\mathbf x^\intercal \boldsymbol\alpha`.
        """
        return dot(x, self.variables().get('effsizes').value)

    def gradient(self, x):
        r"""Gradient of the linear mean function.

        Parameters
        ----------
        x : array_like
            Covariates.

        Returns
        -------
        array_like : :math:`\mathbf x`.
        """
        return dict(effsizes=x)

    @property
    def effsizes(self):
        r"""Effect-sizes parameter."""
        return self.variables().get('effsizes').value

    @effsizes.setter
    def effsizes(self, v):
        self.variables().get('effsizes').value = ascontiguousarray(v)



from __future__ import division

from numpy import clip, errstate, exp, log
from numpy_sugar import epsilon, is_all_finite
from optimix import Function, Scalar, maximize_scalar

from .core import LMMCore
from .scan import FastScanner


class LMM(LMMCore, Function):
    r"""Fast Linear Mixed Models inference via maximum likelihood.

    It perform inference on the model :eq:`lmm1`, explained in the
    :ref:`lmm-intro` section.

    Parameters
    ----------
    y : array_like
        Outcome.
    X : array_like
        Covariates as a two-dimensional array.
    QS : tuple
        Economic eigendecompositon in form of ``((Q0, Q1), S0)`` of a
        covariance matrix ``K``.

    Examples
    --------

    .. doctest::

        >>> from numpy import array
        >>> from numpy_sugar.linalg import economic_qs_linear
        >>> from glimix_core.lmm import LMM
        >>>
        >>> X = array([[1, 2], [3, -1]], float)
        >>> QS = economic_qs_linear(X)
        >>> covariates = array([[1], [1]])
        >>> y = array([-1, 2], float)
        >>> lmm = LMM(y, covariates, QS)
        >>> lmm.fit(verbose=False)
        >>> print('%.3f' % lmm.lml())
        -3.649

    One can also specify which parameters should be fitted:

    .. doctest::

        >>> from numpy import array
        >>> from numpy_sugar.linalg import economic_qs_linear
        >>> from glimix_core.lmm import LMM
        >>>
        >>> X = array([[1, 2], [3, -1]], float)
        >>> QS = economic_qs_linear(X)
        >>> covariates = array([[1], [1]])
        >>> y = array([-1, 2], float)
        >>> lmm = LMM(y, covariates, QS)
        >>> lmm.fix('delta')
        >>> lmm.fix('scale')
        >>> lmm.delta = 0.5
        >>> lmm.scale = 1
        >>> lmm.fit(verbose=False)
        >>> print('%.3f' % lmm.lml())
        -3.832
        >>> lmm.unfix('delta')
        >>> lmm.fit(verbose=False)
        >>> print('%.3f' % lmm.lml())
        -3.713
    """

    def __init__(self, y, X, QS):
        LMMCore.__init__(self, y, X, QS)
        Function.__init__(self, logistic=Scalar(0.0))

        if not is_all_finite(y):
            raise ValueError("There are non-finite values in the phenotype.")

        # self._fix_scale = False
        # self._delta = 0.5
        # self._scale = LMMCore.scale.fget(self)
        self.set_nodata()

    def _get_delta(self):
        v = clip(self.variables().get('logistic').value, -20, 20)
        x = 1 / (1 + exp(-v))
        return clip(x, 1e-5, 1 - 1e-5)

    @property
    def beta(self):
        return LMMCore.beta.fget(self)

    # @beta.setter
    # def beta(self, beta):
    #     LMMCore.beta.fset(self, beta)

    def copy(self):
        r"""Return a copy of this object.

        This is useful for performing new inference based on the results
        of the copied object, as the new LMM object will start its inference
        from the initial solution found in the copied object.

        Returns
        -------
        :class:`.LMM`
            Copy of this object.
        """
        o = LMM.__new__(LMM)

        LMMCore.__init__(o, self._y, self.X, self._QS)
        Function.__init__(
            o, logistic=Scalar(self.variables().get('logistic').value))

        o._fix_scale = self._fix_scale
        # o._delta = self._delta
        o._scale = self._scale

        o.set_nodata()
        return o

    @property
    def delta(self):
        r"""Variance ratio between ``K`` and ``I``."""
        # return self._delta
        v = float(self.variables().get('logistic').value)
        with errstate(over='ignore', under='ignore'):
            v = 1 / (1 + exp(-v))
        return clip(v, epsilon.tiny, 1 - epsilon.tiny)

    @delta.setter
    def delta(self, delta):
        delta = clip(delta, epsilon.tiny, 1 - epsilon.tiny)
        # self._delta = delta
        self.variables().set(dict(logistic=log(delta / (1 - delta))))

    def isfixed(self, var_name):
        r"""Return whether a variable it is fixed or not.

        Parameters
        ----------
        var_name : str
            Possible values are `delta` and `scale`.

        Returns
        -------
        bool
            ``True`` if fixed; ``False`` otherwise.
        """
        if var_name not in ['delta', 'scale']:
            raise ValueError("Possible values are 'delta' and 'scale'.")
        if var_name == 'delta':
            return Function.isfixed(self, 'logistic')
        return self._fix_scale

    @property
    def v0(self):
        r"""First variance.

        Returns
        -------
        float
            :math:`s (1 - \delta)`
        """
        return self.scale * (1 - self.delta)

    @property
    def v1(self):
        r"""Second variance.

        Returns
        -------
        float
            :math:`s \delta`
        """
        return self.scale * self.delta

    def fit(self, verbose=True):
        r"""Maximise the marginal likelihood.

        Parameters
        ----------
        verbose : bool, optional
            ``True`` for progress output; ``False`` otherwise.
            Defaults to ``True``.
        """
        maximize_scalar(self, 'LMM', verbose=verbose)
        self._update()
        self.delta = self._get_delta()

    def fix(self, var_name):
        r"""Disable the optimisation of a given variable.

        Parameters
        ----------
        var_name : str
            Possible values are `delta` and `scale`.
        """
        if var_name not in ['delta', 'scale']:
            raise ValueError("Possible values are 'delta' and 'scale'.")

        if var_name == 'delta':
            super(LMM, self).fix('logistic')
        else:
            if not self._fix_scale:
                self._scale = self.scale
            self._fix_scale = True

    @property
    def fixed_effects_variance(self):
        r"""Variance of the fixed-effects.

        It is defined as the empirical variance of the prior mean.

        Returns
        -------
        array_like
            Estimated variance of the fixed-effects.
        """
        return self.mean.var()

    def get_fast_scanner(self):
        r"""Return :class:`.FastScanner` for association scan.

        Returns
        -------
        :class:`.FastScanner`
            Class initially designed to perform very fast genome-wide scan.
        """
        v0 = self.v0
        v1 = self.v1
        QS = (self._QS[0], v0 * self._QS[1])
        return FastScanner(self._y, self.X, QS, v1)

    def lml(self):
        self.delta = self._get_delta()
        return LMMCore.lml(self)

    @property
    def mean(self):
        r"""Estimated mean :math:`\mathrm X\boldsymbol\beta`.

        Returns
        -------
        array_like
            Mean of the prior.
        """
        return LMMCore.mean.fget(self)

    @property
    def scale(self):
        if self._fix_scale:
            return self._scale
        return LMMCore.scale.fget(self)

    @scale.setter
    def scale(self, v):
        self._scale = v

    def unfix(self, var_name):
        r"""Enable the optimisation of a given variable.

        Parameters
        ----------
        var_name : str
            Possible values are `delta` and `scale`.
        """
        if var_name not in ['delta', 'scale']:
            raise ValueError("Possible values are 'delta' and 'scale'.")
        if var_name == 'delta':
            super(LMM, self).unfix('logistic')
        else:
            self._fix_scale = False

    def value(self):
        self.delta = self._get_delta()
        return self.lml()

    @property
    def X(self):
        return LMMCore.X.fget(self)

    @X.setter
    def X(self, X):
        LMMCore.X.fset(self, X)



from __future__ import absolute_import, division, unicode_literals

from math import fsum

from numpy import dot, isfinite, log, sqrt

from numpy_sugar.linalg import cho_solve, ddot, dotd

from .ep import EP
from .posterior_linear_kernel import PosteriorLinearKernel


def ldot(A, B):
    return ddot(A, B, left=True)


def dotr(A, B):
    return ddot(A, B, left=False)


class EPLinearKernel(EP):
    def __init__(self, nsites):
        super(EPLinearKernel, self).__init__(nsites, PosteriorLinearKernel)

    def lml(self):
        if self._cache['lml'] is not None:
            return self._cache['lml']

        self._update()

        L = self._posterior.L()
        cov = self._posterior.cov
        Q = cov['QS'][0][0]
        S = cov['QS'][1]
        ttau = self._site.tau
        teta = self._site.eta
        ctau = self._cav['tau']
        ceta = self._cav['eta']
        m = self._posterior.mean

        TS = ttau + ctau

        s = cov['scale']
        d = cov['delta']
        A = self._posterior._A
        tQ = sqrt(1 - d) * Q

        lml = [
            -log(L.diagonal()).sum(),
            -0.5 * sum(log(s * S)),
            +0.5 * sum(log(A)),
            # lml += 0.5 * sum(log(ttau)),
            +0.5 * dot(teta * A, dot(tQ, cho_solve(L, dot(tQ.T, teta * A)))),
            -0.5 * dot(teta, teta / TS),
            +dot(m, A * teta) - 0.5 * dot(m, A * ttau * m),
            -0.5 * dot(
                m * A * ttau,
                dot(tQ, cho_solve(L, dot(tQ.T, 2 * A * teta - A * ttau * m)))),
            +sum(self._moments['log_zeroth']),
            +0.5 * sum(log(TS)),
            # lml -= 0.5 * sum(log(ttau)),
            -0.5 * sum(log(ctau)),
            +0.5 * dot(ceta / TS, ttau * ceta / ctau - 2 * teta),
            0.5 * s * d * sum(teta * A * teta)
        ]
        lml = fsum(lml)

        if not isfinite(lml):
            raise ValueError("LML should not be %f." % lml)

        self._cache['lml'] = lml

        return lml

    def lml_derivatives(self, dm):
        self._update()

        L = self._posterior.L()
        ttau = self._site.tau
        teta = self._site.eta
        A = self._posterior._A

        cov = self._posterior.cov
        Q = cov['QS'][0][0]
        S = cov['QS'][1]
        s = cov['scale']
        d = cov['delta']
        tQ = sqrt(1 - d) * Q
        tQS = dotr(tQ, S)
        QS = dotr(Q, S)

        e_m = teta - ttau * self._posterior.mean
        Ae_m = A * e_m
        TA = ttau * A
        tQTAe_m = dot(tQ.T, Ae_m)
        dKAd_m = dot(tQS, tQTAe_m) + d * Ae_m
        w = TA * dot(tQ, cho_solve(L, tQTAe_m))
        QTAe_m = dot(Q.T, Ae_m)
        dKAs_m = -s * dot(QS, QTAe_m) + s * Ae_m
        TAtQ = ldot(TA, tQ)
        LQt = cho_solve(L, Q.T)
        TAQ = ldot(TA, Q)
        r = dotd(TAQ, dot(dot(LQt, TAQ), QS.T)).sum()

        dlml_mean = dot(e_m, ldot(A, dm)) - dot(
            Ae_m, dot(tQ, cho_solve(L, dot(tQ.T, ldot(TA, dm)))))

        dlml_scale = 0.5 * dot(Ae_m, dKAd_m)
        dlml_scale -= sum(w * dKAd_m)
        dlml_scale += 0.5 * dot(w, dot(tQS, dot(tQ.T, w)) + d * w)
        dlml_scale -= 0.5 * dotd(TAtQ, tQS.T).sum()
        dlml_scale -= 0.5 * sum(TA) * d
        dlml_scale += 0.5 * r * (1 - d)**2
        dlml_scale += 0.5 * d * dotd(TAQ, dotr(LQt, TA)).sum() * (1 - d)

        dlml_delta = 0.5 * dot(Ae_m, dKAs_m)
        dlml_delta -= sum(w * dKAs_m)
        dlml_delta += 0.5 * dot(w, -s * dot(QS, dot(Q.T, w)) + s * w)
        dlml_delta += 0.5 * s * dotd(ldot(TA, Q), QS.T).sum()
        dlml_delta -= 0.5 * sum(TA) * s
        dlml_delta -= 0.5 * s * r * (1 - d)
        dlml_delta += 0.5 * s * dotd(TAQ, dotr(LQt, TA)).sum() * (1 - d)

        return dict(mean=dlml_mean, scale=dlml_scale, delta=dlml_delta)



from __future__ import absolute_import, division, unicode_literals

from copy import deepcopy
from math import fsum

from numpy import dot, empty, inf, isfinite, log, maximum, sqrt, zeros
from numpy.linalg import norm
from numpy_sugar import epsilon
from numpy_sugar.linalg import cho_solve, ddot, dotd

from .posterior import Posterior
from .site import Site

MAX_ITERS = 100
RTOL = epsilon.small * 1000
ATOL = epsilon.small


class EP(object):
    r"""Expectation Propagation algorithm.

    Let

    .. math::

        \mathcal N(\mathbf z ~|~ \mathbf m, \mathrm K)

    be the prior distribution.
    This class estimates the log of the marginal likelihood

    .. math::

        p(\mathbf y) = \int \prod_i p(y_i | \mu_i = g(z_i))
            \mathcal N(\mathbf z ~|~ \mathbf m, \mathrm K) \mathrm d\mathbf z

    via Expectation Propagation and provides its gradient.

    Attributes:
        _site (:class:`glimix_core.ep.site.Site`): site-likelihood.
        _psite (:class:`glimix_core.ep.site.Site`): previous
                                                        site-likelihood.
        _cav (:class:`glimix_core.ep.cavity.Cavity`): cavity distribution.
        _posterior (:class:`glimix_core.ep.posterior.Posterior`):
                                                        posterior distribution.
        _moments (dict): moments for KL moment matching.
    """

    def __init__(self, nsites, posterior_type=Posterior):

        self._site = Site(nsites)
        self._psite = Site(nsites)

        self._cav = dict(tau=zeros(nsites), eta=zeros(nsites))
        self._posterior = posterior_type(self._site)

        self._moments = {
            'log_zeroth': empty(nsites),
            'mean': empty(nsites),
            'variance': empty(nsites)
        }

        self._need_update = True
        self._compute_moments = None
        self._cache = dict(lml=None, grad=None)

    def __copy__(self):
        cls = self.__class__
        ep = cls.__new__(cls)

        ep.__dict__['_site'] = deepcopy(self.site)
        ep.__dict__['_psite'] = deepcopy(self.__dict__['_psite'])
        ep.__dict__['_cav'] = deepcopy(self.cav)

        posterior_type = type(self.posterior)
        ep.__dict__['_posterior'] = posterior_type(ep.site)

        ep.__dict__['_moments'] = deepcopy(self.moments)

        ep.__dict__['_need_update'] = self._need_update
        ep.__dict__['_compute_moments'] = None
        ep.__dict__['_cache'] = deepcopy(self._cache)

        return ep

    def _update(self):
        if not self._need_update:
            return

        self._posterior.update()

        i = 0
        tol = -inf
        while i < MAX_ITERS and norm(self._site.tau - self._psite.tau) > tol:
            self._psite.tau[:] = self._site.tau
            self._psite.eta[:] = self._site.eta

            self._cav['tau'][:] = maximum(self._posterior.tau - self._site.tau,
                                          0)
            self._cav['eta'][:] = self._posterior.eta - self._site.eta
            self._compute_moments(self._cav['eta'], self._cav['tau'],
                                  self._moments)

            self._site.update(self._moments['mean'], self._moments['variance'],
                              self._cav)

            self._posterior.update()

            n0 = norm(self._psite.tau)
            n1 = norm(self._cav['tau'])
            tol = RTOL * sqrt(n0 * n1) + ATOL
            i += 1

        if i == MAX_ITERS:
            msg = ('Maximum number of EP iterations has' + ' been attained.')
            msg += " Last EP step was: %.10f." % norm(
                self._site.tau - self._psite.tau)
            raise ValueError(msg)

        self._need_update = False

    @property
    def cav(self):
        return self._cav

    def lml(self):
        if self._cache['lml'] is not None:
            return self._cache['lml']

        self._update()

        L = self._posterior.L()
        Q, S = self._posterior.cov['QS']
        Q = Q[0]
        ttau = self._site.tau
        teta = self._site.eta
        ctau = self._cav['tau']
        ceta = self._cav['eta']
        m = self._posterior.mean

        TS = ttau + ctau

        lml = [
            -log(L.diagonal()).sum(),
            -0.5 * sum(log(S)),
            # lml += 0.5 * sum(log(ttau)),
            +0.5 * dot(teta, dot(Q, cho_solve(L, dot(Q.T, teta)))),
            -0.5 * dot(teta, teta / TS),
            +dot(m, teta) - 0.5 * dot(m, ttau * m),
            -0.5 * dot(m * ttau,
                       dot(Q, cho_solve(L, dot(Q.T, 2 * teta - ttau * m)))),
            +sum(self._moments['log_zeroth']),
            +0.5 * sum(log(TS)),
            # lml -= 0.5 * sum(log(ttau)),
            -0.5 * sum(log(ctau)),
            +0.5 * dot(ceta / TS, ttau * ceta / ctau - 2 * teta)
        ]
        lml = fsum(lml)

        if not isfinite(lml):
            raise ValueError("LML should not be %f." % lml)

        self._cache['lml'] = lml
        return lml

    def lml_derivative_over_cov(self, dQS):
        self._update()

        L = self._posterior.L()
        Q = self._posterior.cov['QS'][0][0]
        ttau = self._site.tau
        teta = self._site.eta

        diff = teta - ttau * self._posterior.mean

        v0 = dot(dQS[0][0], dQS[1] * dot(dQS[0][0].T, diff))
        v1 = ttau * dot(Q, cho_solve(L, dot(Q.T, diff)))
        dlml = 0.5 * dot(diff, v0)
        dlml -= dot(v0, v1)
        dlml += 0.5 * dot(v1, dot(dQS[0][0], dQS[1] * dot(dQS[0][0].T, v1)))
        dqs = ddot(dQS[1], dQS[0][0].T, left=True)
        diag = dotd(dQS[0][0], dqs)
        dlml -= 0.5 * sum(ttau * diag)

        tmp = cho_solve(L, dot(ddot(Q.T, ttau, left=False), dQS[0][0]))
        dlml += 0.5 * sum(ttau * dotd(Q, dot(tmp, dqs)))

        return dlml

    def lml_derivative_over_mean(self, dm):
        self._update()

        L = self._posterior.L()
        Q = self._posterior.cov['QS'][0][0]
        ttau = self._site.tau
        teta = self._site.eta

        diff = teta - ttau * self._posterior.mean

        dlml = dot(diff, dm)
        dlml -= dot(diff, dot(Q, cho_solve(L, dot(Q.T, (ttau * dm.T).T))))

        return dlml

    @property
    def moments(self):
        return self._moments

    @property
    def posterior(self):
        return self._posterior

    def set_compute_moments(self, cm):
        self._compute_moments = cm

    @property
    def site(self):
        return self._site

    def set_prior(self, mean, covariance):
        self._posterior.mean = mean
        self._posterior.cov = covariance
        self._need_update = True
        self._cache['lml'] = None
        self._cache['grad'] = None
